/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type Scsi struct {

	//  Map of attachments, where the key is the integer LUN number on the controller.
	Attachments map[string]Attachment `json:"Attachments,omitempty"`
}
