/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type HvSocket struct {
	Config *HvSocketSystemConfig `json:"Config,omitempty"`

	EnablePowerShellDirect bool `json:"EnablePowerShellDirect,omitempty"`
}
